@@ -1,22 +1,24 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
 import datetime
+import functools
 from decimal import Decimal
 from collections import defaultdict
 
-from sql import Literal, Null
+from sql import Literal, Null, Select
 from sql.aggregate import Max
 from sql.functions import CurrentTimestamp
 from sql.conditionals import Coalesce
 
+from trytond.i18n import gettext
 from trytond.model import ModelSQL, ModelView, fields
-from trytond.wizard import Wizard, StateTransition
-from trytond.pyson import Eval, Or
+from trytond.model.exceptions import AccessError
+from trytond.wizard import Wizard, StateTransition, StateAction
+from trytond.pyson import Eval, Or, PYSONEncoder
 from trytond.transaction import Transaction
 from trytond.pool import Pool, PoolMeta
 from trytond.tools import grouped_slice
 
-from trytond.modules.product import TemplateFunction
 from .move import StockMixin
 
 __all__ = ['Template', 'Product',
@@ -25,13 +27,53 @@ __all__ = ['Template', 'Product',
     'RecomputeCostPrice']
 
 
-class Template:
-    __metaclass__ = PoolMeta
+def check_no_move(func):
+
+    def find_moves(cls, records):
+        pool = Pool()
+        Move = pool.get('stock.move')
+        if cls.__name__ == 'product.template':
+            field = 'product.template'
+        else:
+            field = 'product'
+        for sub_records in grouped_slice(records):
+            moves = Move.search([
+                    (field, 'in', list(map(int, sub_records))),
+                    ],
+                limit=1, order=[])
+            if moves:
+                return moves
+        return False
+
+    @functools.wraps(func)
+    def decorator(cls, *args):
+        pool = Pool()
+        Template = pool.get('product.template')
+        transaction = Transaction()
+        if (transaction.user != 0
+                and transaction.context.get('_check_access')):
+            actions = iter(args)
+            for records, values in zip(actions, actions):
+                for field, msg in Template._modify_no_move:
+                    if field in values:
+                        if find_moves(cls, records):
+                            raise AccessError(gettext(msg))
+                        # No moves for those records
+                        break
+        func(cls, *args)
+    return decorator
+
+
+class Template(metaclass=PoolMeta):
     __name__ = "product.template"
-    quantity = fields.Function(fields.Float('Quantity'), 'sum_product')
-    forecast_quantity = fields.Function(fields.Float('Forecast Quantity'),
-            'sum_product')
-    cost_value = fields.Function(fields.Numeric('Cost Value'),
+    quantity = fields.Function(fields.Float('Quantity',
+        help="The amount of stock in the location."),
+        'sum_product')
+    forecast_quantity = fields.Function(fields.Float('Forecast Quantity',
+        help="The amount of stock expected to be in the location."),
+        'sum_product')
+    cost_value = fields.Function(fields.Numeric('Cost Value',
+        help="The value of the stock in the location."),
         'sum_product')
 
     def sum_product(self, name):
@@ -45,42 +87,18 @@ class Template:
     @classmethod
     def __setup__(cls):
         super(Template, cls).__setup__()
-        cls._error_messages.update({
-                'change_default_uom': ('You cannot change the default uom for '
-                    'a product which is associated to stock moves.'),
-                'change_type': ('You cannot change the type for a product '
-                    'which is associated to stock moves.'),
-                })
         cls.cost_price.states['required'] = Or(
             cls.cost_price.states.get('required', True),
             Eval('type').in_(['goods', 'assets']))
         cls.cost_price.depends.append('type')
         cls._modify_no_move = [
-            ('default_uom', 'change_default_uom'),
-            ('type', 'change_type'),
+            ('default_uom', 'stock.msg_product_change_default_uom'),
+            ('type', 'stock.msg_product_change_type'),
             ]
 
     @classmethod
-    def check_no_move(cls, templates, error):
-        Move = Pool().get('stock.move')
-        for sub_templates in grouped_slice(templates):
-            moves = Move.search([
-                    ('product.template', 'in', [t.id for t in sub_templates]),
-                    ],
-                limit=1, order=[])
-            if moves:
-                cls.raise_user_error(error)
-
-    @classmethod
+    @check_no_move
     def write(cls, *args):
-        if (Transaction().user != 0
-                and Transaction().context.get('_check_access')):
-            actions = iter(args)
-            for templates, values in zip(actions, actions):
-                for field, error in cls._modify_no_move:
-                    if field in values:
-                        cls.check_no_move(templates, error)
-                        break
         super(Template, cls).write(*args)
 
     @classmethod
@@ -92,20 +110,24 @@ class Template:
         Product.recompute_cost_price(products)
 
 
-class Product(StockMixin, object):
-    __metaclass__ = PoolMeta
+class Product(StockMixin, object, metaclass=PoolMeta):
     __name__ = "product.product"
-    quantity = fields.Function(fields.Float('Quantity'), 'get_quantity',
-            searcher='search_quantity')
-    forecast_quantity = fields.Function(fields.Float('Forecast Quantity'),
-            'get_quantity', searcher='search_quantity')
-    cost_value = fields.Function(fields.Numeric('Cost Value'),
+    quantity = fields.Function(fields.Float('Quantity',
+        help="The amount of stock in the location."),
+        'get_quantity', searcher='search_quantity')
+    forecast_quantity = fields.Function(fields.Float('Forecast Quantity',
+        help="The amount of stock expected to be in the location."),
+        'get_quantity', searcher='search_quantity')
+    cost_value = fields.Function(fields.Numeric('Cost Value',
+        help="The value of the stock in the location."),
         'get_cost_value')
 
     @classmethod
     def get_quantity(cls, products, name):
         location_ids = Transaction().context.get('locations')
-        return cls._get_quantity(products, name, location_ids, products)
+        product_ids = list(map(int, products))
+        return cls._get_quantity(
+            products, name, location_ids, grouping_filter=(product_ids,))
 
     @classmethod
     def search_quantity(cls, name, domain=None):
@@ -114,26 +136,32 @@ class Product(StockMixin, object):
 
     @classmethod
     def get_cost_value(cls, products, name):
-        cost_values = {}
+        cost_values = {p.id: None for p in products}
         context = {}
         trans_context = Transaction().context
         if trans_context.get('stock_date_end'):
             # Use the last cost_price of the day
             context['_datetime'] = datetime.datetime.combine(
                 trans_context['stock_date_end'], datetime.time.max)
+            # The date could be before the product creation
+            products = [p for p in products
+                if p.create_date <= context['_datetime']]
         with Transaction().set_context(context):
             for product in cls.browse(products):
-                # The date could be before the product creation
-                if not isinstance(product.cost_price, Decimal):
-                    cost_values[product.id] = None
-                else:
-                    cost_values[product.id] = (Decimal(str(product.quantity))
-                        * product.cost_price)
+                # The product may not have a cost price
+                if product.cost_price is not None:
+                    cost_values[product.id] = (
+                        Decimal(str(product.quantity)) * product.cost_price)
         return cost_values
 
     @classmethod
-    def products_by_location(cls, location_ids, product_ids=None,
-            with_childs=False, grouping=('product',)):
+    @check_no_move
+    def write(cls, *args):
+        super(Product, cls).write(*args)
+
+    @classmethod
+    def products_by_location(cls, location_ids,
+            with_childs=False, grouping=('product',), grouping_filter=None):
         """
         Compute for each location and product the stock quantity in the default
         uom of the product.
@@ -166,7 +194,6 @@ class Product(StockMixin, object):
                     wh_to_add[location.id] = location.storage_location.id
             location_ids = list(location_ids)
 
-        grouping_filter = (product_ids,) + tuple(None for k in grouping[1:])
         query = Move.compute_quantities_query(location_ids, with_childs,
             grouping=grouping, grouping_filter=grouping_filter)
         if query is None:
@@ -175,8 +202,8 @@ class Product(StockMixin, object):
             grouping=grouping, grouping_filter=grouping_filter)
 
         if wh_to_add:
-            for wh, storage in wh_to_add.iteritems():
-                for key in quantities:
+            for wh, storage in wh_to_add.items():
+                for key in list(quantities.keys()):
                     if key[0] == storage:
                         quantities[(wh,) + key[1:]] = quantities[key]
                         if storage in storage_to_remove:
@@ -185,18 +212,7 @@ class Product(StockMixin, object):
 
     @classmethod
     def recompute_cost_price(cls, products):
-        pool = Pool()
-        Template = pool.get('product.template')
-
-        if not isinstance(cls.cost_price, TemplateFunction):
-            digits = cls.cost_price.digits
-            write = cls.write
-            record = lambda p: p
-        else:
-            digits = Template.cost_price.digits
-            write = Template.write
-            record = lambda p: p.template
-
+        digits = cls.cost_price.digits
         costs = defaultdict(list)
         for product in products:
             if product.type == 'service':
@@ -204,40 +220,41 @@ class Product(StockMixin, object):
             cost = getattr(product,
                 'recompute_cost_price_%s' % product.cost_price_method)()
             cost = cost.quantize(Decimal(str(10.0 ** -digits[1])))
-            costs[cost].append(record(product))
+            costs[cost].append(product)
 
         if not costs:
             return
 
         to_write = []
-        for cost, records in costs.iteritems():
+        for cost, records in costs.items():
             to_write.append(records)
             to_write.append({'cost_price': cost})
 
         # Enforce check access for account_stock*
         with Transaction().set_context(_check_access=True):
-            write(*to_write)
+            cls.write(*to_write)
 
     def recompute_cost_price_fixed(self):
         return self.cost_price
 
+    @property
+    def _domain_moves_cost(self):
+        "Returns the domain for moves to use in cost computation"
+        context = Transaction().context
+        return [
+            ('company', '=', context.get('company')),
+            ]
+
     def recompute_cost_price_average(self):
         pool = Pool()
         Move = pool.get('stock.move')
         Currency = pool.get('currency.currency')
         Uom = pool.get('product.uom')
 
-        context = Transaction().context
-
-        if not isinstance(self.__class__.cost_price, TemplateFunction):
-            product_clause = ('product', '=', self.id)
-        else:
-            product_clause = ('product.template', '=', self.template.id)
-
         moves = Move.search([
-                product_clause,
+                ('product', '=', self.id),
                 ('state', '=', 'done'),
-                ('company', '=', context.get('company')),
+                self._domain_moves_cost,
                 ['OR',
                     [
                         ('to_location.type', '=', 'storage'),
@@ -265,10 +282,12 @@ class Product(StockMixin, object):
                         move.company.currency, round=False)
                 unit_price = Uom.compute_price(move.uom, unit_price,
                     self.default_uom)
-                if quantity + qty != 0:
+                if quantity + qty > 0 and quantity >= 0:
                     cost_price = (
                         (cost_price * quantity) + (unit_price * qty)
                         ) / (quantity + qty)
+                elif qty > 0:
+                    cost_price = unit_price
             quantity += qty
         return cost_price
 
@@ -277,10 +296,10 @@ class ProductByLocationContext(ModelView):
     'Product by Location'
     __name__ = 'product.by_location.context'
     forecast_date = fields.Date(
-        'At Date', help=('Allow to compute expected '
-            'stock quantities for this date.\n'
-            '* An empty value is an infinite date in the future.\n'
-            '* A date in the past will provide historical values.'))
+        'At Date',
+        help="The date for which the stock quantity is calculated.\n"
+        "* An empty value calculates as far ahead as possible.\n"
+        "* A date in the past will provide historical values.")
     stock_date_end = fields.Function(fields.Date('At Date'),
         'on_change_with_stock_date_end')
 
@@ -328,34 +347,68 @@ class ProductQuantitiesByWarehouse(ModelSQL, ModelView):
         pool = Pool()
         Move = pool.get('stock.move')
         Location = pool.get('stock.location')
-        move = Move.__table__()
+        Product = pool.get('product.product')
+        Date = pool.get('ir.date')
+        move = from_ = Move.__table__()
+        context = Transaction().context
+        today = Date.today()
 
-        product_id = Transaction().context.get('product')
-        warehouse_id = Transaction().context.get('warehouse', -1)
+        if context.get('product_template') is not None:
+            product = Product.__table__()
+            from_ = move.join(product, condition=move.product == product.id)
+            product_clause = (
+                product.template == context['product_template'])
+        else:
+            product_clause = move.product == context.get('product', -1)
+
+        warehouse_id = context.get('warehouse', -1)
         warehouse_query = Location.search([
                 ('parent', 'child_of', [warehouse_id]),
                 ], query=True, order=[])
-        date_column = Coalesce(move.effective_date, move.planned_date
-            ).as_('date')
-        return move.select(
-            Max(move.id).as_('id'),
-            Literal(0).as_('create_uid'),
-            CurrentTimestamp().as_('create_date'),
-            Literal(None).as_('write_uid'),
-            Literal(None).as_('write_date'),
-            date_column,
-            where=(move.product == product_id)
-            & (move.from_location.in_(warehouse_query)
-                | move.to_location.in_(warehouse_query))
-            & (Coalesce(move.effective_date, move.planned_date) != Null),
-            group_by=(date_column, move.product))
+        date_column = Coalesce(move.effective_date, move.planned_date)
+        return (from_.select(
+                Max(move.id).as_('id'),
+                Literal(0).as_('create_uid'),
+                CurrentTimestamp().as_('create_date'),
+                Literal(None).as_('write_uid'),
+                Literal(None).as_('write_date'),
+                date_column.as_('date'),
+                where=product_clause
+                & (move.from_location.in_(warehouse_query)
+                    | move.to_location.in_(warehouse_query))
+                & (Coalesce(move.effective_date, move.planned_date) != Null)
+                & (date_column != today),
+                group_by=(date_column, move.product))
+            | Select([
+                    Literal(0).as_('id'),
+                    Literal(0).as_('create_uid'),
+                    CurrentTimestamp().as_('create_date'),
+                    Literal(None).as_('write_uid'),
+                    Literal(None).as_('write_date'),
+                    Literal(today).as_('date'),
+                    ]))
 
     @classmethod
     def get_quantity(cls, lines, name):
         Product = Pool().get('product.product')
+        trans_context = Transaction().context
 
-        product_id = Transaction().context.get('product')
-        warehouse_id = Transaction().context.get('warehouse')
+        def valid_context(name):
+            return (trans_context.get(name) is not None
+                and isinstance(trans_context[name], int))
+
+        if not any(map(valid_context, ['product', 'product_template'])):
+            return {l.id: None for l in lines}
+
+        if trans_context.get('product') is not None:
+            grouping = ('product',)
+            grouping_filter = ([trans_context['product']],)
+            key = trans_context['product']
+        else:
+            grouping = ('product.template',)
+            grouping_filter = ([trans_context.get('product_template')],)
+            key = trans_context['product_template']
+        warehouse_id = trans_context.get('warehouse')
 
         dates = sorted(l.date for l in lines)
         quantities = {}
@@ -368,8 +421,10 @@ class ProductQuantitiesByWarehouse(ModelSQL, ModelView):
                 }
             with Transaction().set_context(**context):
                 quantities[date] = Product.products_by_location(
-                    [warehouse_id], [product_id],
-                    with_childs=True).get((warehouse_id, product_id), 0)
+                    [warehouse_id],
+                    grouping=grouping,
+                    grouping_filter=grouping_filter,
+                    with_childs=True).get((warehouse_id, key), 0)
             try:
                 date_start = date + datetime.timedelta(1)
             except OverflowError:
@@ -381,6 +436,32 @@ class ProductQuantitiesByWarehouse(ModelSQL, ModelView):
 
         return dict((l.id, quantities[l.date]) for l in lines)
 
+    @classmethod
+    def get_rec_name(cls, records, name):
+        pool = Pool()
+        Lang = pool.get('ir.lang')
+        Product = pool.get('product.product')
+        Template = pool.get('product.template')
+        Location = pool.get('stock.location')
+        context = Transaction().context
+
+        if context.get('product_template'):
+            name = Template(context['product_template']).rec_name
+        elif context.get('product'):
+            name = Product(context['product']).rec_name
+        else:
+            name = ''
+        if context.get('warehouse'):
+            warehouse_name = Location(context['warehouse']).rec_name
+        else:
+            warehouse_name = '-'
+        lang = Lang.get()
+        names = {}
+        for record in records:
+            names[record.id] = '%s (%s) @ %s' % (
+                name, warehouse_name, lang.strftime(record.date))
+        return names
+
 
 class ProductQuantitiesByWarehouseContext(ModelView):
     'Product Quantities By Warehouse'
@@ -388,16 +469,62 @@ class ProductQuantitiesByWarehouseContext(ModelView):
     warehouse = fields.Many2One('stock.location', 'Warehouse', required=True,
         domain=[
             ('type', '=', 'warehouse'),
-            ])
+            ],
+        help="The warehouse for which the quantities will be calculated.")
 
-    @staticmethod
-    def default_warehouse():
+    @classmethod
+    def default_warehouse(cls):
         Location = Pool().get('stock.location')
-        warehouses = Location.search([
-                ('type', '=', 'warehouse'),
-                ])
-        if len(warehouses) == 1:
-            return warehouses[0].id
+        return Location.get_default_warehouse()
+
+
+class OpenProductQuantitiesByWarehouse(Wizard):
+    "Open Product Quantities By Warehouse"
+    __name__ = 'stock.product_quantities_warehouse.open'
+    start_state = 'open_'
+    open_ = StateAction('stock.act_move_form')
+
+    def do_open_(self, action):
+        pool = Pool()
+        Date = pool.get('ir.date')
+        ProductQuantitiesByWarehouse = pool.get(
+            'stock.product_quantities_warehouse')
+        context = Transaction().context
+        today = Date.today()
+
+        record = ProductQuantitiesByWarehouse(context['active_id'])
+        warehouse_id = context.get('warehouse', -1)
+        domain = [
+            ['OR',
+                [
+                    ('from_location', 'child_of', [warehouse_id], 'parent'),
+                    ('to_location', 'not child_of', [warehouse_id], 'parent'),
+                    ],
+                [
+                    ('to_location', 'child_of', [warehouse_id], 'parent'),
+                    ('from_location', 'not child_of',
+                        [warehouse_id], 'parent'),
+                    ],
+                ],
+            ['OR',
+                ('effective_date', '=', record.date),
+                [
+                    ('effective_date', '=', None),
+                    ('planned_date', '=', record.date),
+                    ],
+                ],
+            ]
+        if record.date < today:
+            domain.append(('state', '=', 'done'))
+        if context.get('product_template'):
+            domain.append(
+                ('product.template', '=', context['product_template']))
+        else:
+            domain.append(('product', '=', context.get('product', -1)))
+        action['pyson_domain'] = PYSONEncoder().encode(domain)
+        action['pyson_search_value'] = None
+        action['name'] += ' (' + record.rec_name + ')'
+        return action, {}
 
 
 class RecomputeCostPrice(Wizard):
